#!/usr/bin/env python
# -*- coding: utf-8 -*-

from functools import partial

import numpy as np

from corna import helpers
from corna.isotopomer import Infopacket
from corna.constants import ISOTOPE_NA_MASS

def na_correct_mimosa_algo_array(parent_frag_m, daughter_frag_n, intensity_m_n, intensity_m_1_n, intensity_m_1_n_1,
                                 isotope, na, decimals):
    '''
    This function calculates the corrected intensity for each and every fragment present in El-MAVEN output file
    
    Input parameters:
    parent_frag_m: Parent formula. Eg. C3H6O7P
    daughter_frag_n: Daughter formula, Eg. PO3
    intensity_m_n, intensity_m_1_n, intensity_m_1_n_1: Intensity of a particular sample for fragments(m_n, m_1_n, m_1_n_1)
    isotope: Contains the isotope information. Eg. C13
    na: naValue for isotopes. Eg. "naValue": {"C12": 0.9889, "C13": 0.0111, "N14": 0.9964, "N15": 0.0036 .... }
        This dictionary is present in element_data.json inside corNA package.
    decimals: decimal points to be considered.
    
    Function variables:
    iso_elem: Gets the element for the isotope. Eg: C13 -> C
    p,d: calculates no. of atoms for iso_elem.
    m,n: calculates no. of atoms in isotope.

    Returns:
    corrected_intensity: Corrected intensity for each sample.
    '''
    iso_elem = helpers.get_isotope_element(isotope)
    p = parent_frag_m.number_of_atoms(iso_elem)
    d = daughter_frag_n.number_of_atoms(iso_elem)
    m = parent_frag_m.get_num_labeled_atoms_isotope(isotope)
    n = daughter_frag_n.get_num_labeled_atoms_isotope(isotope)
    corrected_intensity = intensity_m_n * (1 + na * (p - m)) - intensity_m_1_n * na * ((p - d) - (m - n - 1)) -\
        intensity_m_1_n_1 * na * (d - (n - 1))

    return np.around(corrected_intensity, decimals)


def change_fragment_keys_to_mass(fragments_dict):
    '''
    It converts the keys of fragments_dict to its masses.
    For eg:
    ('2PG 185/79_185.0', '2PG 185/79_79.0') -> (185.0, 79.0)
    
    Note: Only the keys of the dictionary are changed, other than that the dictionary is same.
    '''
    fragment_dict_mass = {}
    for key, value in fragments_dict.items():
        parent_frag, daughter_frag = value.frag
        fragment_dict_mass[
            (parent_frag.isotope_mass, daughter_frag.isotope_mass)] = value
    return fragment_dict_mass


def na_correction_mimosa_by_fragment(fragments_dict, isotope_dict, decimals):
    fragment_dict_mass = change_fragment_keys_to_mass(fragments_dict)
    corrected_dict_mass = {}
    for key, value in fragment_dict_mass.items():
        m_1_n = (key[0] - 1, key[1])
        m_1_n_1 = (key[0] - 1, key[1] - 1)
        parent_frag_m, daughter_frag_n = value.frag
        isotope = parent_frag_m.isotracer
        na = helpers.get_isotope_na(isotope, isotope_dict)
        corrected_data = {}
        for sample_name, intensity_m_n in value.data.items():
            try:
                intensity_m_1_n = fragment_dict_mass[m_1_n].data[sample_name]
            except KeyError:
                intensity_m_1_n = 0
            try:
                intensity_m_1_n_1 = fragment_dict_mass[m_1_n_1].data[sample_name]
            except KeyError:
                intensity_m_1_n_1 = 0
            corrected_data[sample_name] = na_correct_mimosa_algo_array(parent_frag_m,
                                                                       daughter_frag_n, intensity_m_n, intensity_m_1_n,
                                                                       intensity_m_1_n_1, isotope, na, decimals)

        corrected_dict_mass[key] = Infopacket(value.frag,
                                                         corrected_data, value.unlabeled, value.name)
    return corrected_dict_mass


def na_correction_mimosa(metabolite_frag_dict, isotope_dict=ISOTOPE_NA_MASS, decimals=2):
    '''
    na_correction_mimosa method: It is the starting step for processing NA_Correction in MSMS datasets.
    Used in repo: na_correction_shinyapp.

    Parameters:
    metabolite_frag_dict: This dictionary contains information for the following:

        Compound Name: Compound name present in El-MAVEN output file. Eg. [Taurine 128/80, Taurine 124/80, Succinate 120/76 ....]
        Compound Formula: Formula and Parent Formula from metadata_mq file. Eg. [Taurine 124/80: [C2H6NO3S, O3S], Succinate 120/76: [C4H5O4, C3H5O2]] 
        Sample Name: Sample from El-MAVEN output file. 
        intensity value: Peak Area for each different fragment from El-MAVEN output file.
        Different Fragments: It is a key containing different fragments for each and every compound.

        Dict layout:
        {Different Fragments: InfoPacket(frag = [Compound Formula], data = [Sample Name : Intensity Value])}
    
    isotope_dict: Its user independent file. Follow element_data.json inside corNA package. 

        Keys: naValue, element, amu, naturalIsotope, Mass of Isotope, Mass Diff Dict
        naValue: Contains naValue for each element and its isotopes. Eg. ["C12": 0.9889, "C13": 0.0111, "N14": 0.9964, "N15": 0.0036.....]
        element: Contains mapping from isotopes to its parent element. Eg. ["C"  :"C","C12": "C","C13": "C","N":"N","N14": "N", "N15": "N" ....]
        amu: Contains amu value for each element. Eg. ["C12": 12, "C13": 13,"N14": 14, "N15": 15 ...]
        Mass of Isotope: Masses of parent and its isotope. Eg. ["C12":12.0000000000, "C13":13.0033548378, "N14":14.0030740048, "N15":15.0001088982 ...]
        Mass Diff Dict: Difference between isotopes and other parents, isotopes. Eg. ["C13": {"O17": 0.0008622426,"N": 0.0063199444,"H": 0.0029219079,..} ....]

        Dict layout:
        {naValue, element, amu, Mass of Isotope, Mass Diff Dict}
    
    decimals: No. of decimal points we want to consider for results.

    Returns:
    na_corr_dict: Its in the same format of fragments_dict, replaced intensity values with na_corrected values.
    '''
    na_corr_dict = {}
    for metabolite, fragments_dict in metabolite_frag_dict.items():
        na_corr_dict[metabolite] = na_correction_mimosa_by_fragment(fragments_dict, isotope_dict, decimals)

    return na_corr_dict
